@@ -1,22 +1,37 @@
 #!/usr/bin/env python
+
+
 """FGo! - a simple GUI launcher for FlightGear Flight Simulator."""
 
+
+import gettext
+from sys import argv
 from os import chdir
 from Tkinter import Tk
 
 from config import Config
-from gui import App
+from gui.mainwindow import App
+from constants import LOCALE_DIR
+
+
+gettext.install('fgo', LOCALE_DIR, unicode=True)
+
+
+CLI_MESSAGE = """Usage: fgo
+This program does not use command line options. Edit fgo/data/config/presets
+file if you need to run FGo! with some pre-configurations."""
 
 
 def run(working_dir):
     """Initialize application."""
-    # Set current working dirrectory.
+    # Set current working directory.
     chdir(working_dir)
 
     root = Tk()
     root.title('FGo!')
     # Initialize data object.
     data = Config()
+    promptToNotUseCli()
     # Initialize main window.
     app = App(root, data)
 
@@ -31,6 +46,10 @@ def run(working_dir):
 
     root.mainloop()
 
+def promptToNotUseCli():
+    if len(argv) > 1:
+        print _(CLI_MESSAGE)
+
 if __name__ == '__main__':
     from sys import path
 
