@@ -1,11 +1,15 @@
 """This module process data from ~/.fgo folder."""
 
 import os
+import sys
 import gzip
 import gettext
 import codecs
+import traceback
 from Tkinter import IntVar, StringVar
+from tkMessageBox import showerror
 
+from gui.infowindow import InfoWindow
 from constants import *
 
 
@@ -13,22 +17,25 @@ class Config:
 
     """Read/write and store all data from config files."""
 
-    def __init__(self):
+    def __init__(self, master=None):
+        self.master = master
 
-        self.ai_path = ''        # Path to FG_ROOT/AI directory.
-        self.aircraft_dir = ''   # Path to FG_ROOT/Aircraft directory.
-        self.apt_path = ''       # Path to FG_ROOT/Airports/apt.dat.gz file.
-        self.metar_path = ''     # Path to FG_ROOT/Airports/metar.dat.gz file.
+        self.ai_path = ''  # Path to FG_ROOT/AI directory.
+        self.apt_path = ''  # Path to FG_ROOT/Airports/apt.dat.gz file.
+        self.default_aircraft_dir = ''  # Path to FG_ROOT/Aircraft directory.
+        self.metar_path = ''  # Path to FG_ROOT/Airports/metar.dat.gz file.
 
+        self.aircraft_dirs = []  # List of aircraft directories.
         self.aircraft_list = []  # List of aircraft names.
         self.aircraft_path = []  # List of paths to each aircraft.
-        self.airport_icao = []   # List of ICAO codes for each airport.
-        self.airport_name = []   # List of airport names.
-        self.airport_rwy = []    # List of runways present in each airport.
+        self.airport_icao = []  # List of ICAO codes for each airport.
+        self.airport_name = []  # List of airport names.
+        self.airport_rwy = []  # List of runways present in each airport.
+
         self.scenario_list = []  # List of selected scenarios.
         # List of all aircraft carriers found in AI scenario folder.
         # Each entry format is:
-        # ["ship's name", "parking position"... , "scenario's name"]
+        # ["ship name", "parking position"... , "scenario name"]
         self.carrier_list = []
 
         self.settings = []  # List of all settings read from config file.
@@ -37,7 +44,9 @@ class Config:
         self.aircraft = StringVar()
         self.airport = StringVar()
         self.apt_data_source = IntVar()
+        self.auto_update_apt = IntVar()
         self.carrier = StringVar()
+        self.FG_aircraft = StringVar()
         self.FG_bin = StringVar()
         self.FG_root = StringVar()
         self.FG_scenery = StringVar()
@@ -53,36 +62,31 @@ class Config:
         self.TS_scenery = StringVar()
         self.window_geometry = StringVar()
 
-        self.dictionary = {'--aircraft='       : self.aircraft,
-                           '--airport='        : self.airport,
-                           '--fg-root='        : self.FG_root,
-                           '--fg-scenery='     : self.FG_scenery,
-                           '--carrier='        : self.carrier,
-                           '--parkpos='        : self.park,
-                           '--runway='         : self.rwy,
-                           'AI_SCENARIOS='     : self.scenario,
-                           'APT_DATA_SOURCE='  : self.apt_data_source,
-                           'FG_BIN='           : self.FG_bin,
-                           'FG_WORKING_DIR='   : self.FG_working_dir,
-                           # Obsolete keyword, should be removed in the future.
-                           'HOME_DIR='         : self.FG_working_dir,
-                           'FILTER_APT_LIST='  : self.filtredAptList,
-                           'LANG='             : self.language,
-                           'TERRASYNC='        : self.TS,
-                           'TERRASYNC_BIN='    : self.TS_bin,
-                           'TERRASYNC_PORT='   : self.TS_port,
-                           'TERRASYNC_SCENERY=': self.TS_scenery,
-                           'WINDOW_GEOMETRY='  : self.window_geometry}
+        self.keywords = {'--aircraft='       : self.aircraft,
+                         '--airport='        : self.airport,
+                         '--fg-root='        : self.FG_root,
+                         '--fg-scenery='     : self.FG_scenery,
+                         '--carrier='        : self.carrier,
+                         '--parkpos='        : self.park,
+                         '--runway='         : self.rwy,
+                         'AI_SCENARIOS='     : self.scenario,
+                         'APT_DATA_SOURCE='  : self.apt_data_source,
+                         'AUTO_UPDATE_APT='  : self.auto_update_apt,
+                         'FG_BIN='           : self.FG_bin,
+                         'FG_AIRCRAFT='      : self.FG_aircraft,
+                         'FG_WORKING_DIR='   : self.FG_working_dir,
+                         'FILTER_APT_LIST='  : self.filtredAptList,
+                         'LANG='             : self.language,
+                         'TERRASYNC='        : self.TS,
+                         'TERRASYNC_BIN='    : self.TS_bin,
+                         'TERRASYNC_PORT='   : self.TS_port,
+                         'TERRASYNC_SCENERY=': self.TS_scenery,
+                         'WINDOW_GEOMETRY='  : self.window_geometry}
 
         self._createConfDirectory()
         self.update(first_run=True)
 
     def makeInstalledAptList(self):
-#        if self.apt_data_source:
-#            airports = self._findInstalledApt2()
-#        else:
-#            airports = self._findInstalledApt()
-
         airports = self._findInstalledApt()
 
         s = '\n'.join(airports)
@@ -146,7 +150,8 @@ class Config:
         """
         del self.settings
         del self.text
-        del self.aircraft_dir
+        del self.aircraft_dirs
+        del self.default_aircraft_dir
         del self.apt_path
         del self.ai_path
         del self.metar_path
@@ -161,7 +166,9 @@ class Config:
         self.aircraft.set(DEFAULT_AIRCRAFT)
         self.airport.set(DEFAULT_AIRPORT)
         self.apt_data_source.set(0)
+        self.auto_update_apt.set(0)
         self.carrier.set('None')
+        self.FG_aircraft.set('')
         self.FG_bin.set('')
         self.FG_root.set('')
         self.FG_scenery.set('')
@@ -188,16 +195,21 @@ class Config:
                 value = line[cut:]
 
                 if value:
-                    if name in self.dictionary:
+                    if name in self.keywords:
                         if name == 'FILTER_APT_LIST=' and not first_run:
                             pass
                         elif name == 'TERRASYNC=' and not first_run:
                             pass
                         else:
-                            var = self.dictionary[name]
+                            var = self.keywords[name]
                             var.set(value)
 
-        self.aircraft_dir = os.path.join(self.FG_root.get(), AIRCRAFT_DIR)
+        self._setLanguage(self.language.get())
+
+        self.default_aircraft_dir = os.path.join(self.FG_root.get(),
+                                                 DEFAULT_AIRCRAFT_DIR)
+        self.aircraft_dirs = self._readAircraftDirs()
+        self.aircraft_dirs = [self.default_aircraft_dir] + self.aircraft_dirs
 
         self.apt_path = os.path.join(self.FG_root.get(), APT_DAT)
         self.ai_path = os.path.join(self.FG_root.get(), AI_DIR)
@@ -207,9 +219,9 @@ class Config:
         self.scenario_list, self.carrier_list = self._readScenarios()
         self.updateAptLists()
 
-        self._setLanguage(self.language.get())
-
     def updateAptLists(self):
+        if self.auto_update_apt.get() and os.path.exists(self.apt_path):
+            self._autoUpdateApt()
         self.airport_icao, self.airport_name, self.airport_rwy = \
                                                                 self._readApt()
 
@@ -223,7 +235,7 @@ class Config:
         if not path:
             path = CONFIG
         options = []
-        keys = self.dictionary.keys()
+        keys = self.keywords.keys()
         keys.sort()
         # Text
         processed_text = []
@@ -240,7 +252,7 @@ class Config:
                 name = ''
 
             if name in keys:
-                self.dictionary[name].set(value)
+                self.keywords[name].set(value)
             else:
                 if line != CUT_LINE:
                     processed_text.append(line)
@@ -248,11 +260,9 @@ class Config:
         processed_text = '\n'.join(processed_text)
 
         for k in keys:
-            v = self.dictionary[k]
+            v = self.keywords[k]
             if v.get() not in ('Default', 'None'):
-                # Obsolete keyword, should be removed in the future.
-                if k != 'HOME_DIR=':
-                    options.append(k + unicode(v.get()))
+                options.append(k + unicode(v.get()))
 
         s = '\n'.join(options)
 
@@ -261,37 +271,6 @@ class Config:
         config_out.write(processed_text)
         config_out.close()
 
-#    def _findInstalledApt(self):
-#        """Walk thru all scenery, and find installed airports.
-#
-#        Return names of btg.gz files that are shorter than five characters.
-#
-#        """
-#        airport_dict = {}
-#        sceneries = self.FG_scenery.get().split(':')
-#
-#        for scenery in sceneries:
-#            path = os.path.join(scenery, 'Terrain')
-#            if os.path.exists(path):
-#                tree = os.walk(path)
-#
-#                for i in tree:
-#                    if i[2]:
-#                        for file in i[2]:
-#                            file = file.split('.')
-#                            try:
-#                                if len(file[0]) <= 4 and \
-#                                   file[1] == 'btg' and \
-#                                   file[2] == 'gz':
-#                                    airport_dict[file[0]] = None
-#                            except IndexError:
-#                                pass
-#
-#        res = airport_dict.keys()
-#        res.sort()
-#
-#        return res
-
     def _findInstalledApt(self):
         """Walk thru all scenery and find installed airports.
 
@@ -337,58 +316,13 @@ class Config:
 
     def _createConfDirectory(self):
         """Create config directory if non exists."""
-        if not os.path.exists(DATA_DIR):
-            os.mkdir(DATA_DIR)
+        if not os.path.exists(USER_DATA_DIR):
+            os.mkdir(USER_DATA_DIR)
 
-    def _makeApt(self):
+    def _makeApt(self, head=None):
         """Build apt database from apt.dat.gz"""
-        if os.path.exists(self.apt_path):
-            L = []
-            fin = gzip.open(self.apt_path)
-
-            entry = ''
-            lat, lon = 0.0, 0.0
-            runway_count = 0
-            for line in fin:
-                # apt.dat is apparently using iso-8859-1 coding,
-                # so we need to decode it to unicode.
-                line = line.decode('iso-8859-1').encode('utf-8')
-                code = line[:2]
-                # Find ICAO and airport name.
-                if code in ('1 ', '16', '17'):
-                    e = line.strip().split()
-                    name = [e[4], ' '.join(e[5:])]
-                    name = '='.join(name)
-                    entry = name
-                # Find runways.
-                elif code == '10':
-                    e = line.strip().split()
-                    if e[3] != 'xxx':
-                        rwy = e[3]
-                        if rwy.endswith('x'):
-                            rwy = rwy[:-1]
-                        entry += ('=' + rwy)
-                    if e[3] != 'xxx':
-                        lat += float(e[1])
-                        lon += float(e[2])
-                        runway_count += 1
-
-                if line == '\n' and entry:
-                    lat = lat / runway_count
-                    lon = lon / runway_count
-                    coords = ';'.join([str(lat), str(lon)])
-                    entry = '='.join([entry, coords])
-                    L.append(entry)
-
-                    lat, lon = 0.0, 0.0
-                    runway_count = 0
-
-            fin.close()
-            L.sort()
-            fin = open(APT, 'w')
-            for i in L:
-                fin.write(str(i) + '\n')
-            fin.close()
+        if self.FG_root.get():
+            _ProcessApt(self.master, self.apt_path, head)
 
     def _makeText(self):
         """Filter config file entries to be shown in text window."""
@@ -401,7 +335,7 @@ class Config:
                 name = line[:cut]
                 value = line[cut:]
 
-                if name not in self.dictionary:
+                if name not in self.keywords:
                     L.append(name + value)
 
             else:
@@ -421,10 +355,10 @@ class Config:
                 lang_code = gettext.translation(MESSAGES, LOCALE_DIR).info()['language']
             except IOError:
                 lang_code = 'en'
-            path = os.path.join(DEFAULT_CONFIG_DIR, lang_code)
+            path = os.path.join(DEFAULT_CONFIG_DIR, 'config_' + lang_code)
             if not os.path.isfile(path):
                 lang_code = 'en'
-                path = os.path.join(DEFAULT_CONFIG_DIR, lang_code)
+                path = os.path.join(DEFAULT_CONFIG_DIR, 'config_' + lang_code)
             # Load presets if exists.
             try:
                 presets = codecs.open(PRESETS, encoding='utf-8')
@@ -448,25 +382,15 @@ class Config:
             return ['']
 
     def _readAircraft(self):
-        """Walk through Aircraft directory and return two sorted lists:
+        """Walk through Aircraft directories and return two sorted lists:
         list of aircraft names and list of paths to them."""
         try:
             n, p = [], []
-
-            for d in os.listdir(self.aircraft_dir):
-                path = os.path.join(self.aircraft_dir, d)
-                if os.path.isdir(path):
-                    for f in os.listdir(path):
-                        if f[-8:] == '-set.xml':
-                        # Dirty and ugly hack to prevent carrier-set.xml in
-                        # seahawk directory to be attached to the aircraft list.
-                            if path[-7:] == 'seahawk'\
-                            and f == 'carrier-set.xml':
-                                pass
-                            else:
-                                name = f[:-8]
-                                n.append([name.lower(), name, path])
-
+            for dir_ in self.aircraft_dirs:
+                if os.path.isdir(dir_):
+                    self._readAircraftData(dir_, dir_, n)
+                    for d in os.listdir(dir_):
+                        self._readAircraftData(dir_, d, n)
             n.sort()
 
             for i in range(len(n)):
@@ -478,6 +402,21 @@ class Config:
         except OSError:
             return [''], []
 
+    def _readAircraftData(self, dir_, d, n):
+        path = os.path.join(dir_, d)
+        if os.path.isdir(path):
+            for f in os.listdir(path):
+                if f.endswith('-set.xml'):
+                    # Dirty and ugly hack to prevent carrier-set.xml in
+                    # seahawk directory to be attached to the aircraft list.
+                    if (not path.startswith('seahawk') and 
+                        f != 'carrier-set.xml'):
+                        name = f[:-8]
+                        n.append([name.lower(), name, path])
+
+    def _readAircraftDirs(self):
+        return self.FG_aircraft.get().split(':')
+
     def _readApt(self):
         """Read apt list (makes new one if non exists).
 
@@ -504,9 +443,6 @@ class Config:
                         installed_apt.remove(line[0])
                         icao.append(line[0])
                         name.append(line[1])
-#                        # Make runway list if "apt data source"
-#                        # is set to: Standard.
-#                        if not self.apt_data_source.get():
                         rwy_list = []
                         for i in line[2:-1]:
                             rwy_list.append(i)
@@ -518,13 +454,9 @@ class Config:
                     line = line.strip().split('=')
                     icao.append(line[0])
                     name.append(line[1])
-#                    # Make runway list if "apt data source"
-#                    # is set to: Standard.
-#                    if not self.apt_data_source.get():
                     rwy_list = []
                     for i in line[2:-1]:
                         rwy_list.append(i)
-                    rwy_list.sort()
                     rwy.append(rwy_list)
 
             fin.close()
@@ -613,3 +545,270 @@ class Config:
         lon_range = self._calculateRange(lon)
         return lat_range, lon_range
 
+    def _autoUpdateApt(self):
+        if not self.auto_update_apt.get() or not os.path.exists(self.apt_path):
+            return
+        old_timestamp = self._readAptTimestamp()
+        self._updateApt(old_timestamp)
+
+    def _readAptTimestamp(self):
+        if not os.path.exists(APT_TIMESTAMP):
+            self._writeAptTimestamp('')
+        with open(APT_TIMESTAMP) as timestamp:
+            old_modtime = timestamp.read()
+        timestamp.close()
+        return old_modtime
+
+    def _writeAptTimestamp(self, s):
+        with open(APT_TIMESTAMP, 'w') as timestamp:
+            timestamp.write(s)
+        timestamp.close()
+
+    def _getAptModTime(self):
+        return str(os.path.getmtime(self.apt_path))
+
+    def _updateApt(self, old_timestamp):
+        if old_timestamp != self._getAptModTime():
+            self._makeApt(head=_('Modification of apt.dat.gz detected.'))
+            self._writeAptTimestamp(self._getAptModTime())
+
+
+class _ProcessApt:
+
+    """Build apt database from apt.dat.gz"""
+
+    def __init__(self, master, apt_path, head=None):
+        self.master = master
+        self.apt_path = apt_path
+        self.data = []
+        self.main(head)
+
+    def main(self, head):
+        if os.path.exists(self.apt_path):
+            self.make_window(head)
+            try:
+                self.makeApt()
+            except AptdatHeaderError:
+                self.show_aptdat_header_error()
+                self.close_window()
+                return
+            except:
+                print traceback.format_exc()
+                self.show_aptdat_general_error()
+                self.close_window()
+                return
+            self.close_window()
+        else:
+            self.show_no_aptdat_error()
+
+    def make_window(self, head=None):
+        message = _('Generating airport database,\nthis can take a while...')
+        if head:
+            message = '\n'.join((head, message))
+        self.window = InfoWindow(self.master, message, 'Helvetica 12')
+        self.window.update()
+
+    def makeApt(self):
+        self.reset_variables()
+        version = self.get_apt_version_number()
+        self.process_atp(version)
+
+        self.data.sort()
+        self.save_atp_data()
+
+    def reset_variables(self):
+        self.entry = ''
+        self.lat, self.lon = 0.0, 0.0
+        self.runway_count = 0
+
+    def get_apt_version_number(self):
+        """Return version number of apt.dat.gz file.
+        
+        It can be found at the beginning of a second line of the header.
+
+        """
+        with gzip.open(self.apt_path) as fin:
+            origin, number, version = self.read_header(fin)
+            number = self.get_version_number(origin, number, version)
+            return number
+
+    def read_header(self, fin):
+        data = fin.read(24).split()
+        try:
+            origin, number, version = data[0], data[1], data[2]
+            return origin, number, version
+        except:
+            raise AptdatHeaderError
+
+    def get_version_number(self, origin, number, version):
+        if ((origin == 'A' or origin == 'I') and 
+            number.isdigit() and version == 'Version'):
+            return number
+        else:
+            raise AptdatHeaderError
+
+    def process_atp(self, version):
+        with gzip.open(self.apt_path) as fin:
+            for line in fin:
+                # apt.dat is apparently using iso-8859-1 coding,
+                # so we need to decode it to utf-8.
+                line = line.decode('iso-8859-1').encode('utf-8')
+                e = line.strip().split()
+
+                self.process_airport_data(version, e)
+                self.finalize_processing_airport(line)
+
+
+    def process_airport_data(self, version, e):
+        if int(version) < 850:
+            self.process_data_v810(e)
+        else:
+            self.process_data_v850(e)
+
+    def process_data_v810(self, e):
+        if e and e[0] in ('1', '16', '17'):
+            self.get_atp_name(e)
+        # Find runways.
+        elif self.entry and e and e[0] == '10':
+            if e[3] != 'xxx':
+                rwy = e[3]
+                if rwy.endswith('x'):
+                    rwy = rwy[:-1]
+
+                lat, lon = e[1], e[2]
+                self.set_rwy_data(rwy, lat, lon, addotherend=True)
+
+                self.runway_count += 1
+
+    def set_rwy_data(self, rwy, lat, lon, addotherend=False):
+        """Set runway data.
+        
+        If addotherend is set to True, it computes and adds 
+        other end of given rwy. It is needed for apt.dat v 810 
+        and earlier data format, where only one runway end is provided.
+
+        """
+        self.entry += ('=' + rwy)
+        if addotherend:
+            self.add_other_end(rwy)
+        self.set_position_data(lat, lon)
+
+    def set_position_data(self, lat, lon):
+        self.lat += float(lat)
+        self.lon += float(lon)
+
+    def add_other_end(self, rwy):
+        rwy_number = self.compute_other_end(rwy)
+        if rwy_number:
+            self.entry += ('=' + rwy_number)
+
+    def compute_other_end(self, rwy):
+        if not rwy.startswith('H'):
+            prefix = self.compute_heading(rwy)
+            suffix = self.change_left_right(rwy)
+            return prefix + suffix
+
+    def compute_heading(self, rwy):
+        number = self.get_heading(rwy)
+        if number <= 36 and number > 18:
+            prefix = str(number - 18)
+        elif number > 0 and number <= 18:
+            prefix = str(number + 18)
+        return prefix
+
+    def get_heading(self, rwy):
+        while not rwy.isdigit():
+            rwy = rwy[:-1]
+        return int(rwy)
+
+    def change_left_right(self, rwy):
+        suffix = ''
+        if rwy.endswith('L'):
+            suffix = 'R'
+        elif rwy.endswith('R'):
+            suffix = 'L'
+        elif rwy.endswith('C'):
+            suffix = 'C'
+        return suffix
+
+    def process_data_v850(self, e):
+        if e and e[0] in ('1', '16', '17'):
+            self.get_atp_name(e)
+        # Find runways.
+        elif self.entry and e and e[0] == '100':
+            self.set_first_rwy_v850(e)
+            self.set_second_rwy_v850(e)
+            self.runway_count += 2
+        # Find water runways.
+        elif self.entry and e and e[0] == '101':
+            self.set_first_water_rwy_v850(e)
+            self.set_second_water_rwy_v850(e)
+            self.runway_count += 2
+        # Find helipads.
+        elif self.entry and e and e[0] == '102':
+            self.set_helipad_v850(e)
+            self.runway_count += 1
+
+    def set_first_rwy_v850(self, e):
+        self.set_rwy_data(e[8], e[9], e[10])
+        
+    def set_second_rwy_v850(self, e):
+        self.set_rwy_data(e[17], e[18], e[19])
+
+    def set_first_water_rwy_v850(self, e):
+        self.set_rwy_data(e[3], e[4], e[5])
+
+    def set_second_water_rwy_v850(self, e):
+        self.set_rwy_data(e[6], e[7], e[8])
+
+    def set_helipad_v850(self, e):
+        self.set_rwy_data(e[1], e[2], e[3])
+
+    def get_atp_name(self, e):
+        """Get ICAO and airport name."""
+        name = [e[4], ' '.join(e[5:])]
+        name = '='.join(name)
+        self.entry = name
+
+    def finalize_processing_airport(self, line):
+        if self.entry and (line in ['\n', '\r\n'] or line.startswith('99')):
+            self.averaged_coordinates()
+            self.set_coordinates()
+            self.reset_variables()
+
+    def averaged_coordinates(self):
+        if self.runway_count:
+            self.lat /= self.runway_count
+            self.lon /= self.runway_count
+
+    def set_coordinates(self):
+        coords = ';'.join([str(self.lat), str(self.lon)])
+        self.entry = '='.join([self.entry, coords])
+        self.data.append(self.entry)
+
+    def save_atp_data(self):
+        with open(APT, 'w') as fin:
+            for i in self.data:
+                fin.write(str(i) + '\n')
+
+    def show_aptdat_header_error(self):
+        message = _('Cannot read version number of apt.dat database.\n\n'
+                    'FGo! expects a proper header information at the '
+                    'beginning of apt.dat. For details about data format see '
+                    'http://data.x-plane.com/designers.html#Formats')
+        showerror(_('Error'), message)
+
+    def show_aptdat_general_error(self):
+        message = _('Cannot read apt.dat database.')
+        showerror(_('Error'), message)
+
+    def show_no_aptdat_error(self):
+        message = _('Cannot find apt.dat database.')
+        showerror(_('Error'), message)
+
+    def close_window(self):
+        self.window.destroy()
+
+
+class AptdatHeaderError(Exception):
+    pass
